from __future__ import unicode_literals, print_function

import sys
import re
import os
import csv
from datetime import datetime

import click
from prompt_toolkit import Application, CommandLineInterface, AbortAction
from prompt_toolkit.buffer import AcceptAction
from prompt_toolkit.history import FileHistory
from prompt_toolkit.shortcuts import create_prompt_layout, create_eventloop
from prompt_toolkit.filters import Always
from prompt_toolkit.styles import PygmentsStyle
from pygments.token import Token
from py2neo.error import Unauthorized
from py2neo.packages.httpstream import SocketError, http

from cycli import __version__
from cycli.lexer import CypherLexer
from cycli.style import CypherStyle
from cycli.completer import CypherCompleter
from cycli.buffer import CypherBuffer
from cycli.binder import CypherBinder
from cycli.neo4j import Neo4j
from cycli.table import pretty_print_table
from cycli.cypher import Cypher

cypher = Cypher()


def get_tokens(x):
        return [(Token.Prompt, "> ")]


class Cycli:

    def __init__(self, host, port, username, password, logfile, filename, ssl, read_only):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.logfile = logfile
        self.filename = filename
        self.ssl = ssl
        self.read_only = read_only

    def write_to_logfile(self, query, response):
        results = response["results"]
        duration = response["duration"]
        error = response["error"]

        self.logfile.write("> {}\n".format(query))
        self.logfile.write("{}\n".format(results))

        if not error:
            self.logfile.write("{} ms\n\n".format(duration))

    @staticmethod
    def write_to_csvfile(data):
        filename = "cycli {}.csv".format(datetime.now().strftime("%Y-%m-%d at %I.%M.%S %p"))

        with open(filename, "wt") as csvfile:
            csvwriter = csv.writer(csvfile)
            csvwriter.writerow(data.columns)

            for row in data:
                csvwriter.writerow(row)

        csvfile.close()

    def run(self):
        neo4j = Neo4j(self.host, self.port, self.username, self.password, self.ssl)
        neo4j.connect()
        self.neo4j = neo4j

        try:
            labels = neo4j.labels()
            relationship_types = neo4j.relationship_types()
            properties = neo4j.properties()

        except Unauthorized:
            print("Unauthorized. See cycli --help for authorization instructions.")
            return

        except SocketError:
            print("Connection refused. Is Neo4j turned on?")
            return

        if self.filename:
            queries = self.filename.read()
            queries = queries.split(";")[:-1]

            for query in queries:
                query += ";"
                query = query.strip()

                print("> " + query)
                self.handle_query(query)
                print()

            return

        click.secho(" ______     __  __     ______     __         __    ", fg="red")
        click.secho("/\  ___\   /\ \_\ \   /\  ___\   /\ \       /\ \   ", fg="yellow")
        click.secho("\ \ \____  \ \____ \  \ \ \____  \ \ \____  \ \ \  ", fg="green")
        click.secho(" \ \_____\  \/\_____\  \ \_____\  \ \_____\  \ \_\ ", fg="blue")
        click.secho("  \/_____/   \/_____/   \/_____/   \/_____/   \/_/ ", fg="magenta")

        print("\nVersion: {}".format(__version__))
        print("Bug reports: https://github.com/nicolewhite/cycli/issues\n")

        completer = CypherCompleter(labels, relationship_types, properties)

        layout = create_prompt_layout(
            lexer=CypherLexer,
            get_prompt_tokens=get_tokens,
            reserve_space_for_menu=8,
        )

        buff = CypherBuffer(
            accept_action=AcceptAction.RETURN_DOCUMENT,
            history=FileHistory(filename=os.path.expanduser('~/.cycli_history')),
            completer=completer,
            complete_while_typing=True,
        )

        application = Application(
            style=PygmentsStyle(CypherStyle),
            buffer=buff,
            layout=layout,
            on_exit=AbortAction.RAISE_EXCEPTION,
            key_bindings_registry=CypherBinder.registry
        )

        cli = CommandLineInterface(application=application, eventloop=create_eventloop())

        try:
            while True:
                document = cli.run()
                query = document.text
                self.handle_query(query)

        except Exception:
            print("Goodbye!")

    def handle_query(self, query):
        run_n = re.match('run-([0-9]+) (.*)', query, re.DOTALL)
        save_csv = query.startswith("save-csv ")

        if cypher.is_a_write_query(query) and self.read_only:
            print("Query aborted. You are in read-only mode.")

        elif query in ["quit", "exit"]:
            raise Exception

        elif query == "help":
            print_help()

        elif query == "refresh":
            self.neo4j.refresh()

        elif query == "schema":
            self.neo4j.print_schema()

        elif query == "schema-indexes":
            self.neo4j.print_indexes()

        elif query == "schema-constraints":
            self.neo4j.print_constraints()

        elif query == "schema-labels":
            self.neo4j.print_labels()

        elif query == "schema-rels":
            self.neo4j.print_relationship_types()

        elif query.startswith("env"):
            if query == "env":
                for key, value in self.neo4j.parameters.items():
                    print("{0}={1}".format(key, value))
            else:
                key = query[3:]
                key = key.strip("'\"[]")
                value = self.neo4j.parameters.get(key)

                if value is not None:
                    print(value)

        elif query.startswith("export "):
            if "=" not in query:
                print("Set parameters with export key=value.")
            else:
                params = query.replace("export ", "").strip()
                key, value = params.split("=", 1)
                key = key.strip()
                value = value.strip()

                try:
                    value = eval(value)
                    self.neo4j.update_parameters(key, value)
                except Exception as e:
                    print(e)

        else:
            count = int(run_n.group(1)) if run_n else 1
            query = run_n.group(2) if run_n else query
            query = query[len("save-csv "):] if save_csv else query

            if count <= 0 or not query:
                raise Exception

            total_duration = 0
            index = 0
            error = False

            while index < count:
                response = self.neo4j.cypher(query)

                results = response["results"]
                duration = response["duration"]
                error = response["error"]

                print(results)

                if not error:
                    ms = "Run {}: {} ms\n".format(index + 1, duration) if run_n else "{} ms".format(duration)
                    print(ms)

                if self.logfile:
                    self.write_to_logfile(query, response)

                if save_csv and not error:
                    self.write_to_csvfile(results[0])

                total_duration += duration
                index += 1

            if run_n and not error:
                print("Total duration: {} ms".format(total_duration))


def print_help():
    headers = ["Keyword", "Description"]

    rows = [
        ["quit", "Exit cycli."],
        ["exit", "Exit cycli."],
        ["help", "Display this text."],
        ["refresh", "Refresh schema cache."],
        ["run-n", "Run a Cypher query n times."],
        ["export", "Set a parameter with export key=value."],
        ["save-csv", "Save the query results to a CSV file."],
        ["schema", "Display indexes, constraints, labels, and relationship types."],
        ["schema-indexes", "Display indexes."],
        ["schema-constraints", "Display constraints."],
        ["schema-labels", "Display labels."],
        ["schema-rels", "Display relationship types."],
        ["CTRL-D", "Exit cycli if the input is blank."],
        ["CTRL-C", "Abort and rollback the currently-running query."]
    ]

    pretty_print_table(headers, rows)


@click.command()
@click.option("-v", "--version", is_flag=True, help="Show cycli version and exit.")
@click.option("-h", "--host", default="localhost", help="The host address of Neo4j.")
@click.option("-P", "--port", default="7474", help="The port number on which Neo4j is listening.")
@click.option("-u", "--username", help="Username for Neo4j authentication.")
@click.option("-p", "--password", help="Password for Neo4j authentication.")
@click.option("-t", "--timeout", help="Set a global socket timeout for queries.", type=click.INT)
@click.option('-l', '--logfile', type=click.File(mode="a", encoding="utf-8"), help="Log every query and its results to a file.")
@click.option("-f", "--filename", type=click.File(mode="rb"), help="Execute semicolon-separated Cypher queries from a file.")
@click.option("-s", "--ssl", is_flag=True, help="Use the HTTPS protocol.")
@click.option("-r", "--read-only", is_flag=True, help="Do not allow any write queries.")
def run(host, port, username, version, timeout, password, logfile, filename, ssl, read_only):
    if version:
        print("cycli {}".format(__version__))
        sys.exit(0)

    if username and not password:
        password = click.prompt("Password", hide_input=True, show_default=False, type=str)

    if timeout:
        http.socket_timeout = timeout

    cycli = Cycli(host, port, username, password, logfile, filename, ssl, read_only)
    cycli.run()


if __name__ == '__main__':
    run()



from pygments.style import Style
from pygments.token import Token

class CypherStyle(Style):
    styles = {
        Token.Keyword: '#3498DB',
        Token.String: '#E67E22',
        Token.Name: '#1ABC9C',
        Token.Pattern: '#E74C3C',

        Token.Line: 'bg:#000000 #ffffff',

        Token.LineNumber:  'bg:#ffffaa #000000',
        Token.Menu.Completions.Completion.Current: 'bg:#00aaaa #000000',
        Token.Menu.Completions.Completion:         'bg:#008888 #ffffff',
        Token.Menu.Completions.ProgressButton:     'bg:#003333',
        Token.Menu.Completions.ProgressBar:        'bg:#00aaaa',

        Token.Toolbar.Completions:  'bg:#888800 #000000',
        Token.Toolbar.Completions.Arrow: 'bg:#888800 #000000',
        Token.Toolbar.Completions.Completion:  'bg:#aaaa00 #000000',
        Token.Toolbar.Completions.Completion.Current:  'bg:#ffffaa #000000 bold',

        Token.AfterInput: 'bg:#ff44ff #000000',
    }



import requests
from py2neo import Graph, authenticate, Resource
from datetime import datetime
from cycli.table import pretty_print_table


class Neo4j:

    __relationship_types = None
    __labels = None
    __constraints = None
    __indexes = None

    parameters = {}

    def __init__(self, host, port, username=None, password=None, ssl=False):
        self.username = username
        self.password = password

        self.host_port = "{host}:{port}".format(host=host, port=port)
        self.url = "{protocol}{host_port}/db/data/".format(protocol="https://" if ssl else "http://",
                                                           host_port=self.host_port)

    def connect(self):
        if self.username and self.password:
            authenticate(self.host_port, self.username, self.password)

        graph = Graph(self.url)
        self.graph = graph

    def cypher(self, query):
        error = False
        start = datetime.now()
        tx = self.graph.cypher.begin()

        try:
            tx.append(query, parameters=self.parameters)
            results = tx.process()
            tx.commit()
        except Exception as e:
            results = e
            error = True
        except KeyboardInterrupt:
            tx.rollback()
            results = ""
            error = True

        end = datetime.now()
        duration = int(round((end - start).total_seconds() * 1000))

        return {"results": results, "duration": duration, "error": error}

    def labels(self):
        if not self.__labels:
            labels_resource = Resource(self.graph.uri.string + "labels")
            self.__labels = sorted(list(frozenset(labels_resource.get().content)))

        return self.__labels

    def relationship_types(self):
        if not self.__relationship_types:
            relationship_types_resource = Resource(self.graph.uri.string + "relationship/types")
            self.__relationship_types = sorted(list(frozenset(relationship_types_resource.get().content)))

        return self.__relationship_types

    def constraints(self):
        if not self.__constraints:
            constraints = []
            constraints_resource = Resource(self.graph.uri.string + "schema/constraint")
            constraints_content = list(constraints_resource.get().content)

            for i in constraints_content:
                constraint = dict(i)
                constraint["property_keys"] = list(constraint["property_keys"])
                constraints.append(constraint)

            self.__constraints = constraints

        return self.__constraints

    def indexes(self):
        if not self.__indexes:
            indexes = []
            for label in self.labels():
                index_resource = Resource(self.graph.uri.string + "schema/index/" + label)
                indexes_content = list(index_resource.get().content)

                for i in indexes_content:
                    index = dict(i)
                    index["property_keys"] = list(index["property_keys"])
                    indexes.append(index)

            self.__indexes = indexes

        return self.__indexes

    def update_parameters(self, key, value):
        self.parameters[key] = value

    def refresh(self):
        self.__labels = None
        self.__relationship_types = None
        self.__indexes = None
        self.__constraints = None
        self.labels()
        self.relationship_types()
        self.indexes()
        self.constraints()

    def print_labels(self):
        headers = ["Labels"]
        rows = [[x] for x in self.labels()]

        pretty_print_table(headers, rows)

    def print_relationship_types(self):
        headers = ["Relationship Types"]
        rows = [[x] for x in self.relationship_types()]

        pretty_print_table(headers, rows)

    def print_constraints(self):
        headers = ["Constraints"]
        constraints = self.constraints()
        rows = [[x] for x in self.format_constraints_indexes(constraints)]

        pretty_print_table(headers, rows)

    def print_indexes(self):
        headers = ["Indexes"]
        indexes = self.indexes()
        rows = [[x] for x in self.format_constraints_indexes(indexes)]

        pretty_print_table(headers, rows)

    def format_constraints_indexes(self, values):
        return [":{}({})".format(value["label"], ",".join(value["property_keys"])) for value in values]

    def print_schema(self):
        headers = ["Labels", "Relationship Types", "Constraints", "Indexes"]

        columns = [self.labels()[:]]
        columns.append(self.relationship_types()[:])
        columns.append(self.format_constraints_indexes(self.constraints()[:]))
        columns.append(self.format_constraints_indexes(self.indexes()[:]))

        max_length = len(max(columns, key=len))
        [x.extend([""] * (max_length - len(x))) for x in columns]
        rows = [[x[i] for x in columns] for i in range(max_length)]

        pretty_print_table(headers, rows)

    def properties(self):
        url = self.url + "propertykeys"
        r = requests.get(url, auth=(self.username, self.password))
        props = r.json()
        return sorted(props)


from __future__ import unicode_literals
from prompt_toolkit.completion import Completer, Completion
from cycli.cypher import Cypher


cypher = Cypher()


class CypherCompleter(Completer):

    def __init__(self, labels, relationship_types, properties):
        super(CypherCompleter, self).__init__()

        self.labels = labels
        self.relationship_types = relationship_types
        self.properties = properties

    def get_completions(self, document, complete_event):
        chars_before_cursor = document.get_word_before_cursor(WORD=True)
        all_text = document.text_before_cursor
        lookup = self.find_last_alphabetic_chunk(chars_before_cursor)
        choices = []

        if not chars_before_cursor or self.unclosed_strings(all_text):
            return

        if self.most_recent_non_alpha(chars_before_cursor) == ":":
            if self.looking_for(chars_before_cursor) == "label":
                choices = self.labels
            elif self.looking_for(chars_before_cursor) == "relationship":
                choices = self.relationship_types
            else:
                return
        elif self.most_recent_non_alpha(chars_before_cursor) == ".":
            loc = chars_before_cursor.find(".")

            if loc:
                identifier = chars_before_cursor[:loc]

                if identifier.isalnum() and any(c.isalpha() for c in identifier):
                    choices = self.properties
                else:
                    return
        elif chars_before_cursor[-1].isalpha():
            last_cypher_word = self.most_recent_cypher_word(all_text)
            choices = cypher.most_probable_next_keyword(last_cypher_word)
        else:
            return

        completions = self.find_matches(lookup, choices)

        for completion in completions:
            yield Completion(completion, -len(lookup))

    def find_matches(self, word, choices):
        word = word.lower()
        lower_choices = [x.lower() for x in choices]

        completions = []

        for i, choice in enumerate(lower_choices):
            if choice.startswith(word):
                completions.append(choices[i])

        return completions

    def looking_for(self, chars_before_cursor):
        paren = chars_before_cursor.rfind("(")
        bracket = chars_before_cursor.rfind("[")
        curly = chars_before_cursor.rfind("{")

        d = dict(label=paren, relationship=bracket, curly=curly)
        return max(d, key=d.get)

    def find_last_alphabetic_chunk(self, chars):
        chars = list(chars)
        chars.reverse()

        keep = []

        for c in chars:
            if c.isalpha():
                keep.append(c)
            else:
                break

        keep.reverse()
        return "".join(keep)

    def most_recent_cypher_word(self, all_text):
        text = " " + all_text
        keyword_indices = [(word, text.rfind(" " + word + " ")) for word in cypher.KEYWORDS]
        function_indices = [(word, text.rfind(" " + word + "(")) for word in cypher.FUNCTIONS]

        indices = keyword_indices + function_indices

        # If no keywords were found, we want to be in the "" state in the Markov model.
        if not any([i[1] > -1 for i in indices]):
            return ""

        most_recent = max(indices, key=lambda i:i[1])[0]
        return most_recent

    def most_recent_non_alpha(self, chars):
        chars = list(chars)
        chars.reverse()

        for c in chars:
            if not c.isalpha():
                return c

        return ""

    def unclosed_strings(self, chars):
        return chars.count("\"") % 2 != 0 or chars.count("'") % 2 != 0


from prompt_toolkit.buffer import Buffer
from prompt_toolkit.filters import Condition

class CypherBuffer(Buffer):

    def __init__(self, *args, **kwargs):

        @Condition
        def is_multiline():
            text = self.document.text
            return not self.user_wants_out(text)

        super(self.__class__, self).__init__(*args, is_multiline=is_multiline, **kwargs)

    def user_wants_out(self, text):
        return any(
            [
                text.endswith(";"),
                text.endswith("\n"),
                text == "quit",
                text == "exit",
                text == "help",
                text == "refresh",
                text == "schema",
                text == "schema-constraints",
                text == "schema-indexes",
                text == "schema-labels",
                text == "schema-rels",
                text.startswith("env"),
                text.startswith("export ")
            ]
        )


from cycli.markov import markov
import re

class Cypher:

    def __init__(self):
        self.FUNCTIONS = [
            "ABS",
            "ACOS",
            "ALL",
            "ALLSHORTESTPATHS",
            "ANY",
            "ASIN",
            "ATAN",
            "ATAN2",
            "AVG",
            "CEIL",
            "COALESCE",
            "COLLECT",
            "COS",
            "COT",
            "COUNT",
            "DEGREES",
            "E",
            "ENDNODE",
            "EXISTS",
            "EXP",
            "EXTRACT",
            "FILTER",
            "FLOOR",
            "HAS",
            "HAVERSIN",
            "HEAD",
            "ID",
            "KEYS",
            "LABELS",
            "LAST",
            "LEFT",
            "LENGTH",
            "LOG",
            "LOG10",
            "LOWER",
            "LTRIM",
            "MAX",
            "MIN",
            "NODE",
            "NODES",
            "NONE",
            "PERCENTILECONT",
            "PERCENTILEDISC",
            "PI",
            "RADIANS",
            "RAND",
            "RANGE",
            "REDUCE",
            "REL",
            "RELATIONSHIP",
            "RELATIONSHIPS",
            "REPLACE",
            "RIGHT",
            "ROUND",
            "RTRIM",
            "SHORTESTPATH",
            "SIGN",
            "SIN",
            "SINGLE",
            "SIZE",
            "SPLIT",
            "SQRT",
            "STARTNODE",
            "STDEV",
            "STDEVP",
            "STR",
            "SUBSTRING",
            "SUM",
            "TAIL",
            "TAN",
            "TIMESTAMP",
            "TOFLOAT",
            "TOINT",
            "TRIM",
            "TYPE",
            "UPPER"
        ]

        self.KEYWORDS = [
            "AND",
            "AS",
            "ASC",
            "ASCENDING",
            "ASSERT",
            "BY",
            "CASE",
            "COMMIT",
            "CONSTRAINT",
            "CONTAINS",
            "CREATE",
            "CSV",
            "CYPHER",
            "DELETE",
            "DESC",
            "DESCENDING",
            "DETACH",
            "DISTINCT",
            "DROP",
            "ELSE",
            "END",
            "ENDS",
            "EXPLAIN",
            "FALSE",
            "FIELDTERMINATOR",
            "FOREACH",
            "FROM",
            "HEADERS",
            "IN",
            "INDEX",
            "IS",
            "LIMIT",
            "LOAD",
            "MATCH",
            "MERGE",
            "NOT",
            "NULL",
            "ON",
            "OPTIONAL",
            "OR",
            "ORDER",
            "PERIODIC",
            "PROFILE",
            "REMOVE",
            "RETURN",
            "SCAN",
            "SET",
            "SKIP",
            "START",
            "STARTS",
            "THEN",
            "TRUE",
            "UNION",
            "UNIQUE",
            "UNWIND",
            "USING",
            "WHEN",
            "WHERE",
            "WITH",
            "XOR"
        ]

        self.markov = markov

    def words(self):
        return sorted(self.FUNCTIONS + self.KEYWORDS)

    def most_probable_next_keyword(self, current_keyword):
        return [x[0] for x in self.markov[current_keyword]]

    def is_a_write_query(self, query):
        query = query.upper()
        update_words = ["CREATE", "MERGE", "DELETE", "SET", "REMOVE", "DROP"]

        unquoted = re.findall('(?:^|"|\'|`)([^"|\'|`]*)(?:$|"|\'|`)', query, flags=re.DOTALL)
        unquoted = " " + "".join(unquoted) + " "
        unquoted = unquoted.replace("\n", " ")
        unquoted = unquoted.replace("\r", " ")
        unquoted = unquoted.replace("\t", " ")

        sightings = [unquoted.find(" " + word + " ") for word in update_words]

        return any([x > -1 for x in sightings])